@@ -80,10 +80,30 @@ setMethod("fft", signature(z="gpu.matrix.tensorflow"), function(z){
   return(res)
 })
 
-setMethod("xtfrm", signature(x="gpu.matrix.tensorflow"), function(x){
-  return(as.numeric(x))
-})
 
+
+# setMethod("sort", signature(x="gpu.matrix.tensorflow", decreasing = "missing"), function(x,decreasing){
+#   if (x@sparse) {
+#     res <- as.vector(sort(x@gm$values))
+#   }else{
+#     res <- as.vector(tensorflow::tf$sort(tensorflow::tf$reshape(x@gm,length(x))))
+#   }
+#   return(res)
+# })
+#
+# setMethod("sort", signature(x="gpu.matrix.tensorflow", decreasing = "logical"), function(x,decreasing){
+#   if(decreasing){
+#     decreasing="DESCENDING"
+#   }else{
+#     decreasing="ASCENDING"
+#     }
+#   if (x@sparse) {
+#     res <- as.vector(tensorflow::tf$sort(x@gm$values, direction = decreasing))
+#   }else{
+#     res <- as.vector(tensorflow::tf$sort(tensorflow::tf$reshape(x@gm,length(x)),direction=decreasing))
+#   }
+#   return(res)
+# })
 setMethod("sort", signature(x="gpu.matrix.tensorflow"), function(x,decreasing=F,...){
   if(decreasing){
     decreasing="DESCENDING"
@@ -230,7 +250,7 @@ setMethod("cbind2",signature(x = "gpu.matrix.tensorflow", y = "ANY"), function(x
   }else{
     if (x@sparse) x <- warningSparseTensor(x)
     if (y@sparse) y <- warningSparseTensor(y)
-    res <- gpu.matrix.tensorflow(cbind(x@gm,y@gm))
+    res <- gpu.matrix.tensorflow(tensorflow::tf$concat(c(x@gm,y@gm),axis=1L))
   }
 
   if (is.null(colnames(x)) & !is.null(colnames(y))) colnames(x) <- rep(NA,ncol(x))
@@ -262,7 +282,7 @@ setMethod("cbind2",signature(x = "ANY", y = "gpu.matrix.tensorflow"), function(x
   }else{
     if (x@sparse) x <- warningSparseTensor(x)
     if (y@sparse) y <- warningSparseTensor(y)
-    res <- gpu.matrix.tensorflow(cbind(x@gm,y@gm))
+    res <- gpu.matrix.tensorflow(tensorflow::tf$concat(c(x@gm,y@gm),axis=1L))
   }
 
   if (is.null(colnames(x)) & !is.null(colnames(y)) & !is.vector(xOrigin)) colnames(x) <- rep(NA,ncol(x))
@@ -292,7 +312,7 @@ setMethod("rbind2", signature(x = "gpu.matrix.tensorflow", y = "ANY"), function(
   }else{
     if (x@sparse) x <- warningSparseTensor(x)
     if (y@sparse) y <- warningSparseTensor(y)
-    res <- gpu.matrix.tensorflow(rbind(x@gm,y@gm))
+    res <- gpu.matrix.tensorflow(tensorflow::tf$concat(c(x@gm,y@gm),axis=0L))
   }
 
 
@@ -319,7 +339,7 @@ setMethod("rbind2",signature(x = "ANY", y = "gpu.matrix.tensorflow"), function(x
   }else{
     if (x@sparse) x <- warningSparseTensor(x)
     if (y@sparse) y <- warningSparseTensor(y)
-    res <- gpu.matrix.tensorflow(rbind(x@gm,y@gm))
+    res <- gpu.matrix.tensorflow(tensorflow::tf$concat(c(x@gm,y@gm),axis=0L))
   }
 
 
@@ -404,32 +424,6 @@ setMethod("tcrossprod", signature(x = "gpu.matrix.tensorflow", y = "missing"), f
 
 
 
-# setMethod("outer", signature(X = "gpu.matrix.tensorflow", Y = "ANY"), function(X,Y, ...){
-#
-#   castMatrix <- castTypeOperations(X,Y)
-#   X <- castMatrix[[1]]
-#   Y <- castMatrix[[2]]
-#
-#   return(as.array(tensorflow::tf$tensordot(X@gm, Y@gm, axes=0L)))
-#
-# } )
-# setMethod("outer", signature(X = "ANY", Y = "gpu.matrix.tensorflow"), function(X,Y, ...){
-#
-#   castMatrix <- castTypeOperations(X,Y)
-#   X <- castMatrix[[1]]
-#   Y <- castMatrix[[2]]
-#
-#   return(as.array(tensorflow::tf$tensordot(X@gm, Y@gm, axes=0L)))
-#
-# } )
-#
-# setMethod("%o%", signature(X = "gpu.matrix.tensorflow", Y = "ANY"), function(X,Y){
-#   return(outer(X,Y))
-# })
-# setMethod("%o%", signature(X = "ANY", Y = "gpu.matrix.tensorflow"), function(X,Y){
-#   return(outer(X,Y))
-# })
-
 tf_kron <- function(X,Y){
   castMatrix <- castTypeOperations(X,Y, sameType = T)
   X <- castMatrix[[1]]
@@ -520,22 +514,66 @@ setMethod("solve", signature(a = "ANY", b = "gpu.matrix.tensorflow"), function(a
   return(res)
 })
 
-setMethod("qr", signature(x="gpu.matrix.tensorflow"), function(x){
-
+setMethod("qr", signature(x="gpu.matrix.tensorflow"), function(x,...){
   x <- warningSparseTensor(x)
-  res <- qr(x@gm)
-  res$qr <- gpu.matrix.tensorflow(res$qr)
-
+  qrTF <- tensorflow::tf$linalg$qr(x@gm,full_matrices = TRUE)
+  res <- list(q=gpu.matrix.tensorflow(qrTF[0]), r=gpu.matrix.tensorflow(qrTF[1]), x=x)
   return(res)
 })
 
-# setGeneric("rankMatrix", function(x) standardGeneric("rankMatrix"))
-# setMethod("rankMatrix", signature(x="gpu.matrix.tensorflow"), function(x){
-#   x <- warningSparseTensor(x)
-#   res <- Matrix::rankMatrix(x@gm)
+setMethod("qr.solve", signature(a="gpu.matrix.tensorflow", b="gpu.matrix.tensorflow"), function(a,b){
+  b <- warningInteger(b)
+  castMatrix <- castTypeOperations(a,b,sameType = T)
+  a <- castMatrix[[1]]
+  b <- castMatrix[[2]]
+  qr_gpu <- qr(a)
+  # res_solve <- tensorflow::tf$linalg$triangular_solve(qr_gpu$r@gm, (t(qr_gpu$q) %*% b)@gm, lower = F)
+  # res <- gpu.matrix.tensorflow(res_solve)
+  res <- qr.solve(a=qr_gpu,b=b)
+  return(res)
+})
+
+
+setMethod("qr.solve", signature(a="gpu.matrix.tensorflow", b="ANY"), function(a,b){
+  b <- warningInteger(b)
+  castMatrix <- castTypeOperations(a,b,sameType = T)
+  a <- castMatrix[[1]]
+  b <- castMatrix[[2]]
+  qr_gpu <- qr(a)
+  # res_solve <- tensorflow::tf$linalg$triangular_solve(qr_gpu$r@gm, (t(qr_gpu$q) %*% b)@gm, lower = F)
+  # res <- gpu.matrix.tensorflow(res_solve)
+  res <- qr.solve(a=qr_gpu,b=b)
+  return(res)
+})
+
+setMethod("qr.solve", signature(a="ANY", b="gpu.matrix.tensorflow"), function(a,b){
+  b <- warningInteger(b)
+  castMatrix <- castTypeOperations(a,b,sameType = T)
+  a <- castMatrix[[1]]
+  b <- castMatrix[[2]]
+  qr_gpu <- qr(a)
+  # res_solve <- tensorflow::tf$linalg$triangular_solve(qr_gpu$r@gm, (t(qr_gpu$q) %*% b)@gm, lower = F)
+  # res <- gpu.matrix.tensorflow(res_solve)
+  res <- qr.solve(a=qr_gpu,b=b)
+  return(res)
+})
+
+
+
+# setMethod("qr.solve", signature(a="list", b="ANY"), function(a,b){
+#
+  # castMatrix <- castTypeOperations(a[[1]], b, sameType = T)
+  # b <- castMatrix[[2]]
+  #
+  # qr_gpu <- a
+  # res_solve <- tensorflow::tf$linalg$triangular_solve(qr_gpu$r@gm, (t(qr_gpu$q) %*% b)@gm, lower = F)
+  # res <- gpu.matrix.tensorflow(res_solve)
 #   return(res)
 # })
 
+
+
+
 #Se debe mejorar
 setMethod("eigen", signature(x="gpu.matrix.tensorflow"), function(x){
 
@@ -634,6 +672,15 @@ setMethod("rowMeans", signature(x = "gpu.matrix.tensorflow"), function(x){
 
   return(res)
 })
+setGeneric("sum2", function(x) standardGeneric("sum2"))
+setMethod("sum2", signature(x = "gpu.matrix.tensorflow"), function(x){
+  if (x@sparse) {
+    res <- as.numeric(tensorflow::tf$sparse$reduce_sum(x@gm))
+  }else{
+    res <- as.numeric(tensorflow::tf$reduce_sum(x@gm))
+  }
+  return(res)
+})
 setMethod("sum", signature(x = "gpu.matrix.tensorflow"), function(x){
   if (x@sparse) {
     res <- as.numeric(tensorflow::tf$sparse$reduce_sum(x@gm))
@@ -768,11 +815,7 @@ setMethod("which.min", signature(x = "gpu.matrix.tensorflow"), function(x){
   return(res)
 })
 
-setMethod("aperm", signature(a="gpu.matrix.tensorflow"), function(a,perm,...){
-  res <- aperm(as.matrix(a),perm)
 
-  return(res)
-})
 
 
 
@@ -818,7 +861,7 @@ applyTest <- function (X, MARGIN, FUN, ..., simplify = TRUE)
     return(if (is.null(ans)) ans else if (length(d.ans) <
                                           2L) ans[1L][-1L] else array(ans, d.ans, dn.ans))
   }
-  newX <- aperm(X, c(s.call, s.ans))
+  newX <- t(X)
   dim(newX) <- c(prod(d.call), d2)
   ans <- vector("list", d2)
   if (length(d.call) < 2L) {
@@ -874,7 +917,7 @@ applyTest <- function (X, MARGIN, FUN, ..., simplify = TRUE)
 
 # setGeneric("apply", function(X, MARGIN, FUN, ..., simplify = TRUE) standardGeneric("apply"))
 setMethod("apply", signature(X="gpu.matrix.tensorflow"), function(X, MARGIN, FUN, ..., simplify = TRUE){
-  applyTest(X, MARGIN, FUN, ..., simplify = TRUE)
+  applyTest(X, MARGIN, FUN, ..., simplify = simplify)
 
 })
 
@@ -889,7 +932,7 @@ setMethod("cov", signature(x = "gpu.matrix.tensorflow", y = "ANY"), function(x,y
   x_ <- t(x) - colMeans(x)
   y_ <- t(y) - colMeans(y)
 
-  res <- tcrossprod(x_, y_)/(ncol(x)-1)
+  res <- tcrossprod(x_, y_)/(ncol(x_)-1)
   return(res)
 })
 
@@ -903,14 +946,15 @@ setMethod("cov", signature(x = "ANY", y = "gpu.matrix.tensorflow"), function(x,y
   x_ <- t(x) - colMeans(x)
   y_ <- t(y) - colMeans(y)
 
-  res <- tcrossprod(x_, y_)/(ncol(x)-1)
+  res <- tcrossprod(x_, y_)/(ncol(x_)-1)
   return(res)
 })
 
 setMethod("cov", signature(x = "gpu.matrix.tensorflow", y = "missing"), function(x,y){
   x <- warningInteger(x)
   x <- warningSparseTensor(x)
-  res <- tcrossprod(t(x) - colMeans(x))/(ncol(x)-1)
+  x_ <- t(x) - colMeans(x)
+  res <- tcrossprod(x_/(ncol(x_)-1))
   return(res)
 })
 
